"""
Data module within the Brigit package.

Contains several different data structures that refer to relevant
information required by other modules within the package.

Copyright by Raúl Fernández Díaz
"""
import os
import json


def read_stats() -> dict:
    # Define paths to stats
    current_dir = os.path.dirname(__file__)
    stats_path = os.path.join(current_dir, 'stats')
    residues_path = os.path.join(stats_path, 'residue_statistics.json')
    gaussian_path = os.path.join(stats_path, 'gaussian_statistics.json')

    # Read stats
    with open(residues_path) as reader_1:
        stats = json.load(reader_1)
    with open(gaussian_path) as reader_2:
        gaussian_stats = json.load(reader_2)

    return stats, gaussian_stats


ALL_CHANNELS = ['hydrophobic', 'aromatic', 'hbond_acceptor', 'hbond_donor',
                'positive_ionizable', 'negative_ionizable', 'metal',
                'excluded_volume']
ALL_METALS = {
    'LI', 'BE',
    'NA', 'MG', 'V', 'CR', 'MN', 'FE', 'CO', 'NI', 'CU', 'ZN',
    'K',  'CA',      'MO',       'RU', 'RH', 'PD', 'AG', 'CD',
    'RB', 'SR',                        'IR', 'PT', 'AU', 'HG',
    'CS', 'BA'
}
ATOMIC_MASSES = {
    'H': 1.01,
    'D': 2.01,
    'C': 12.01,
    'N': 14.01,
    'N1+': 14.01,
    'N1-': 14.01,
    'O': 16.0,
    'O1+': 16.0,
    'O1-': 16.0,
    'O-1': 16.0,
    'O2-': 16.0,
    'S': 32.06,
    'S2-': 32.06,
    'P': 30.97,
    'P1+': 30.97,
    'SE': 78.96,
}
CHANNELS = ['hydrophobic', 'hbond_acceptor', 'hbond_donor',
            'positive_ionizable', 'negative_ionizable', 'excluded_volume']


CHANNELS_DICT = {channel: idx for idx, channel in enumerate(ALL_CHANNELS)}

METAL_RESNAMES = {
    'DDH', 'BEF', 'H79', '1PT', 'AV2', 'P9G', 'SNF', 'CFC', 'HCB', 'BF4',
    'CAS', 'TIL', 'A72', 'IR3', 'HDE', '2J0', 'RHX', 'MYQ', 'PBM', 'SQ1',
    '72B', 'HEG', 'CON', '0UE', 'CPO', 'ALB', 'RKP', 'VER', 'F4S', 'FMI',
    'DRB', '118', 'HEM', '3UQ', 'COB', 'PLL', 'ZPT', 'S31', 'CPT', 'CNC',
    'HEB', 'IUM', 'JSE', 'B1M', 'PCU', 'HIF', 'S32', 'KSB', 'CVC', 'HRU',
    'PCL', '1MK', 'RHD', 'AC9', 'WO3', 'HEO', 'V', 'PT4', 'H58', 'F3S',
    'AM', 'ALF', 'DTZ', '31Q', 'HCO', 'M43', '0JC', 'WO5', 'PT', 'FCE',
    'CU1', 'R7U', 'CU', 'AMW', '0TE', '8WV', 'DVW', 'B9F', 'A71', 'I83',
    'PC4', 'CBY', 'RKL', 'CUL', 'EU3', 'DHE', 'TTO', 'MD9', 'PC3', 'RKM',
    '51O', 'ELJ', 'CL7', 'WO2', 'AF3', '6CQ', 'BVA', 'N7H', 'RH3', 'CUS',
    'JSD', 'PD', 'RE', 'HEV', 'TBY', 'C4R', 'QHL', 'XAX', '2GO', 'MN',
    'DAZ', 'MM1', 'CS', 'MM4', 'HES', 'OS', '73M', 'MM6', 'CAC', 'HBF',
    'SM', 'NCP', 'CA', 'MOS', 'BVQ', 'HG', '7G4', 'HGB', 'IRI', 'ART',
    'ZEM', 'APW', 'PR', 'RUA', 'R6A', 'HNN', 'FES', 'FDD', '35N', 'DW2',
    '5IR', 'GBF', '522', 'OBV', 'H57', 'U1', 'CL1', 'PMR', 'CM2', 'PB',
    'C7P', 'NI', 'HAS', 'EMT', 'A6R', 'GB0', 'ZN9', '3T3', 'MGF', 'VN4',
    'REJ', 'CE', 'IMF', 'YT3', 'COY', 'CSR', 'BE7', 'BCB', '4IR', 'FEC',
    'CAD', 'UVC', 'PHG', '2FH', 'CMH', 'GA', '0TN', 'UFE', 'AUF', 'BTF',
    'TB', 'LOS', 'RIR', '0H2', 'AU3', 'NTE', '6CO', 'CLA', 'BA', 'BAZ',
    'CL0', 'SR', 'AIV', 'MBO', '89R', 'OMO', 'OS1', 'SF3', 'FLL', '4A6',
    'LA', 'IR', 'TPT', 'DW5', '3CG', 'PA0', 'DY', 'HE6', 'EFE', '6B6',
    '9RU', 'ZN7', 'HDD', '5AU', 'CCH', 'RU1', 'CO', 'HEA', 'COH', 'AUC',
    '4KV', 'F43', 'SI4', 'VN3', 'HEC', 'RPS', 'I42', 'ME3', 'HDM', 'R1Z',
    'HME', 'FC6', 'IME', 'J1R', 'GD3', '2PT', '3WB', '6BP', 'RSW', 'POR',
    '9QB', 'RUX', 'ZNH', 'AG1', 'OSV', 'OFE', 'CR', '4MO', 'DWC', 'CAF',
    'PFC', '9D7', '8AR', 'FE', 'MNR', 'OS4', 'B13', 'NA', '3NI', 'RHE',
    'RTA', 'WO4', 'GIX', 'B12', 'CD', 'ZCM', 'CM1', 'RU8', 'ARS', 'BOZ',
    'CUP', '4HE', 'V5A', '9ZQ', 'C5L', '08T', 'SIR', 'YPT', 'YB2', 'MOO',
    'OHX', '7GE', 'AVC', 'MP1', 'LI', 'SF4', 'B22', 'RFB', 'FCI', 'ER3',
    'BF2', 'RUR', 'B30', 'MTQ', 'RU2', '6ZJ', 'GD', 'DEF', '3Q7', 'DW1',
    'MM5', 'YBT', 'YOK', 'BCL', 'ASR', 'I2A', 'SI7', 'Y1', 'CHL', 'RH',
    '6O0', 'SI8', 'N2R', 'PTN', 'REI', 'B1R', 'ITM', 'MOM', 'SI0', 'SVP',
    'RMD', 'MG', 'T0M', 'FE9', '76R', 'RHM', '2MO', 'FS1', 'SI9', '4PU',
    'VVO', '4TI', 'FE2', 'TL', 'UNL', 'CFQ', '2T8', 'RBN', 'MH2', 'VEA',
    'NOB', 'T1A', 'RU', 'MMC', '8CY', 'RKF', 'MTV', 'PL1', '6HE', 'ISW',
    'HE5', 'PMB', 'SB', 'RU7', 'REP', 'RTB', 'ZN6', 'GCR', 'REQ', 'L2D',
    'AL', '9Q8', '1Y8', 'HFM', 'RTC', 'R1C', 'CF', 'MH0', 'CSB', 'L4D',
    'L2M', 'AU', 'AOH', 'AG', 'ZND', 'DRU', 'KYT', 'TTA', 'CX8', 'EU',
    'ATS', '6WO', 'NRU', 'HG2', 'TCN', 'REO', 'QPT', 'SMO', 'N2W', '3ZZ',
    'DEU', 'YOL', 'KYS', 'NXC', 'LU', 'SBO', 'ZN8', 'EMC', 'HKL', 'R4A',
    'MM2', '7BU', '0OD', 'JM1', '6MO', 'YXX', 'JSC', 'HGD', 'SKZ', 'BS3',
    'FEL', 'PNI', 'HB1', 'RHL', 'RML', 'ZN0', 'TH', 'ZN', 'MSS', 'YB',
    'TA0', 'N2N', 'RBU', 'DOZ', 'NCO', 'K', 'RXO', 'YOM', 'FDE', 'LPT',
    '188', 'RAX', 'BJ5', 'CU6', 'RUI', 'IN', 'SXC', 'OEY', 'LSI', 'CX3',
    'PCD', 'RFA', 'ICA', 'B6F', 'MN3', 'MNH', 'RUC', 'M10', 'MOW', '11R',
    'HNI', '7MT', 'W', 'HGI', '1FH', 'WPC', 'MAP', 'R9A', 'ZN5', '68G',
    'RUH', 'RCZ', 'E52', 'HO3', 'MO', '7HE', 'VO4', 'FEM', 'T9T', 'CB5',
    'HO', '6WF', '3CO', 'CLN', '5LN', 'RB', 'TAS', 'CQ4', 'B1Z'
}

NEW_METALS = {f'{metal}{idx}' for idx in range(1, 4) for metal in ALL_METALS}
NEW_METALS = NEW_METALS.union(ALL_METALS)

RESNAME2LETTER_DICT = {
    'VAL': 'V', 'ILE': 'I', 'LEU': 'L', 'GLU': 'E', 'GLN': 'Q',
    'ASP': 'D', 'ASN': 'N', 'HIS': 'H', 'TRP': 'W', 'PHE': 'F',
    'TYR': 'Y', 'ARG': 'R', 'LYS': 'K', 'SER': 'S', 'THR': 'T',
    'MET': 'M', 'ALA': 'A', 'GLY': 'G', 'PRO': 'P', 'CYS': 'C'
}


if __name__ == '__main__':
    print('File with different data stored to be used as global variables.')



"""
Main BrigitMetalPredictor module.

Contains 1 class:
    - Brigit

Copyright by Raúl Fernández Díaz
"""
import multiprocessing
import os
import time
import warnings

import numpy as np
import torch
from moleculekit.molecule import Molecule
from moleculekit.tools.atomtyper import prepareProteinForAtomtyping
from moleculekit.tools.voxeldescriptors import getCenters, getVoxelDescriptors
from sklearn.cluster import Birch
from tqdm import tqdm

from .utils.data import CHANNELS
from .utils.pdb_parser import protein
from .utils.scoring import (coordination_scorer, discrete_score,
                            gaussian_score, parse_residues)
from .utils.tools import (distribute, find_coordinators,
                          get_undesired_channels, load_model, ordered_list,
                          select_desired_channels, set_up_cuda)


class Brigit():
    """
    Main class encapsulating the necessary attributes and methods
    for predicting protein-metal binding regions.
    """
    def __init__(
        self,
        device: str = 'cuda',
        device_id: int = 0,
        model: str = 'BrigitCNN',
        **kwargs
    ):
        """
        Initialize a new instance of the class.
        Set CUDA up if any GPU device is available.
        """
        if device == 'cuda':
            if torch.cuda.is_available():
                self.device = f'{device}:{device_id}'
                set_up_cuda(device_id)
            else:
                warnings.warn('Warning: No CUDA compatible GPU was available,\
                    defaulting to CPU.')
                self.device = 'cpu'
        else:
            self.device = device
            if torch.cuda.is_available():
                warnings.warn('Using CPU for caluclations, although there is\
                    a CUDA compatible GPU available. Using the GPU would\
                    accelerate significantly the calculations.')
        self.model = load_model(self.device)

    def cnn_evaluation(
        self,
        target: str,
        cnn_threshold: float = 0.5,
        verbose: bool = True,
        stride: int = 1,
        pH: float = 7.4,
        **kwargs
    ) -> np.array:
        """
        Read a PDB file either from a local path or from the database. Voxelise
        it and then evaluate its points using the CNN model.

        Args:
            target (str): PDB ID or path to local PDB file.
            cnn_threshold (float): Threshold to consider a certain point as
                positive, i.e., metal-bindingin. Default: 0.5.
            verbose (bool): How much information should be displayed.
                Default: True.
            stride (int): Movement of the sliding window through the voxelised
                representation of the protein. Default: 1.
            pH (float): Medium acidity. Default: 7.4.

        Returns:
            scores (np.array): Matrix with 4 columns where the first 3
                determine the coordinates of the points and the 4th the score
                assigned to that location.
        """
        if not isinstance(target, str):
            message = 'Target has to be:\n  a) string with PDB ID or\n'
            message += '  b) path to local PDB file.'
            raise TypeError(message)

        # Voxelize the protein
        if verbose:
            print(f'Voxelizing target: {target}', end='\n\n')

        vox, p_centers, nvoxels = self.voxelize(target, **kwargs)

        # CNN evaluation
        if verbose:
            print(f'\nCNN evaluation of target: {target}', end='\n\n')

        scores = self.cnn(vox, p_centers, stride, **kwargs)

        return scores

    def predict(
        self,
        target: str,
        metals: list,
        max_coordinators: int = 4,
        outputfile: str = None,
        cnn_threshold: float = 0.5,
        combined_threshold: float = 0.5,
        verbose: int = 1,
        cluster_radius: float = 5.0,
        **kwargs
    ) -> np.array:
        """
        Main method of the class. It takes a protein target and it
        predicts the suitability of its regions for coordinating metals.

        First, the protein is voxelized using the Moleculekit library. 6
        channels are computed: hidrophobicity, H-bond acceptance, H-bond
        reception, positive ionizability, negative ionizability, excluded
        volume. The voxelized representation is parsed with the CNN and each
        point is evaluated. Then, all points with scores superior to the
        threshold are further analysed. In this case, their relative position
        to key atoms in the protein backbone will provide a statistical score
        that serves as indicator as to how well coordinated a hypothetical
        metal ion will be.

        The resulting points will be clustered attending to their spatial
        coordinates and the centers of the clusters will be assigned as
        a weighted mean of the coordinates of all points within the cluster.
        The weights will be the square of the scores. So, that the center is
        closer to highly coordinable regions.

        Finally, two output files will be generated:
            1. A '_brigit.pdb' file which contains all probes with a score
                superior to `combined_threshold`.
            2. A '.clusters' file which contains all residues within
                `cluster_radius` of every cluster center. Thus, summarising
                the most likely coordinating residues in the protein. Each
                cluster will have assigned a global score which will be a sum
                of the scores of all the probes that comprise it.

        Args:
            target (str): Path to the PDB file or PDB code of the target
                protein.
            metals (list): List of IUPAC symbols for the metals to be
                evaluated. There is a mode, 'GENERIC', that uses statistics
                built from all metals simultaneously.
            max_coordinators (int, optional): Maximum number of coordinators
                expected. Value affects the sensibility of the coordination
                analysis. Defaults to 4.
            outputfile (str, optional): Path where the output files are to
                be stored. It is recommended that it does not contain tags,
                e.g., '.pdb', as the program itself will asign them.
                Defaults to None.
            cnn_threshold (float, optional): Score threshold for the output of
                the CNN model to be further analysed. Defaults to 0.5.
            combined_threshold (float, optional): Final threshold to consider a
                region. Defaults to 0.5.
            verbose (int, optional): Ammount of information to be displayed
                during computation. Defaults to 1.
            cluster_radius (float, optional): Radius of the clusters used to
                organise the results. Affects mainly the final summary.
                Defaults to 5.0.

        Raises:
            TypeError: In case the target is neither a PDB code nor a Path to a
                local PDB file.

        Returns:
            np.array: Set of coordinates and their score.
        """
        start = time.time()
        verbose = bool(verbose)

        scores = self.cnn_evaluation(
            target,
            cnn_threshold,
            **kwargs
        )

        # Coordination analysis
        if verbose:
            print(f'\nCoordination analysis of target: {target}', end='\n\n')

        for metal in metals:
            coor_scores, molecule, coordinators = self.coordination_analysis(
                target, metal.upper(), scores, max_coordinators, cnn_threshold,
                verbose, **kwargs
            )

            best_scores = self.filter_best_results(
                coor_scores, combined_threshold
            )

            if len(best_scores) < 1:
                print('No positions are predicted to accomodate metals.')
                end = time.time()
                if verbose:
                    print(f'Computation took {round(end-start, 2)} s.', end='\n\n')
                return scores


            if verbose:
                print('Clusterizing results', end='\n\n')

            # Spatial clusterization of best positions
            centers, cluster_scores = self.clusterize(
                best_scores, molecule, cluster_radius
            )

            if verbose:
                print('Preparing and writing output files', end='\n\n')

            # Preparing and writing output files
            if outputfile is None:
                if target.endswith('.pdb'):
                    new_outputfile = os.path.split(target)[1].split('.')[0]
                    new_outputfile = f'{new_outputfile}_{metal.lower()}'

                else:
                    new_outputfile = f'{target}_{metal.lower()}'
            else:
                new_outputfile = f'{outputfile}_{metal.lower()}'

            try:
                a = kwargs['args']
            except KeyError:
                kwargs['args'] = None

            self.print_clusters(
                centers, molecule, new_outputfile, coordinators,
                cluster_radius, kwargs['args']
            )
            self.print_PDB(
                target, new_outputfile, best_scores, combined_threshold,
                cluster_scores, molecule
            )

        end = time.time()

        if verbose:
            print(f'Computation took {round(end-start, 2)} s.', end='\n\n')
        return scores

    def coordination_analysis(
        self,
        target: protein,
        metal: str,
        scores: np.array,
        max_coordinators: int = 2,
        threshold: float = 0.5,
        verbose: bool = True,
        residue_score: str = 'gaussian',
        cnn_weight: float = 0.5,
        residues: int = 7,
        threads: int = 8,
        **kwargs
    ) -> tuple:
        """
        It uses a series of statistical values regarding backbone
        preorganization of metal-binding amino acids, to evaluate
        how well coordinated each probe would be.

        There are 2 main functions that can be used for this scoring:
            1. 'discrete': which uses the median values of the statistics
                and evaluates whether a certain geometric requirements are
                fulfilled or not.
            2. 'gaussian': which instead of using a boolean argument, scores
                more positively those regions as they are closer to the actual
                central values in a continuos fashion.

        More details on these functions can be found in the Docstrings for
        the module `utils/scoring`.

        Args:
            target (protein): Path to the PDB file or PDB code of the target
                protein.
            max_coordinators (int): Maximum number of coordinators
                expected. Value affects the sensibility of the coordination
                analysis.
            metal (str): IUPAC symbol of the metal to be evaluated. There is a
                mode, 'General', that uses statistics built from all metals
                simultaneously.
            scores (np.array): Array with dimensions (len(probes), 4) where the
                first 3 dimensions describe the cartesian coordinates occupied
                by the probe, and the fourth is its score.
            threshold (float): (CNN) score value that determines whether the
                coordination of a certain probe will be analysed or not.
            verbose (bool): Ammount of information to be displayed
                during computation.
            residue_score (str): Scoring function used to evaluate the
                coordination through side chains.
            cnn_weight (float): Proportion of the final score that will be
                informed by the CNN. Its complementary (1 - cnn_weight)
                corresponds to the proportion of the final score informed
                by the coordination analysis.
            residues (int): Number of the most likely residue coordinators for
                any specific metal.
            threads (int): Number of threads to be used during the
                voxelization.

        Raises:
            IndexError: Occurs in the event that the residue coordination
                scoring function selected is not implemented.
            IndexError: Occurs in the event that the backbone coordination
                scoring function selected is not implemented.

        Returns:
            tuple: (np.array, protein) the np.array corresponds to
                the updated scores array with the cartesian coordiantes
                of the probes and their scores.
        """
        zeros = np.zeros(np.shape(scores[:, 3]))
        molecule = protein(target, True)
        coordinators, stats, gaussian_stats = find_coordinators(
            metal, residues
        )
        molecule_info = parse_residues(molecule, coordinators)

        if residue_score == 'discrete':
            residue_score = discrete_score
        elif residue_score == 'gaussian':
            residue_score = gaussian_score
        else:
            raise IndexError(
                f'Residue coordination scoring function: {residue_score} not\
                    currently implemented.'
            )

        chunks = distribute(scores, threads)
        args = [(molecule, chunks[idx], stats, gaussian_stats, molecule_info,
                coordinators, residue_score, max_coordinators,
                threshold, cnn_weight) for idx in range(threads)]
        pool = multiprocessing.Pool(threads)
        zeros = pool.starmap(analyse_probes, args)
        new_zeros = []
        for element in zeros:
            for subelement in element:
                new_zeros.append(subelement)

        new_zeros = np.array(new_zeros)

        return new_zeros, molecule, coordinators

    def voxelize(
        self,
        target: str,
        voxelsize: float = 1.0,
        buffer: int = 6,
        validitychecks: bool = False,
        channels: list = CHANNELS,
        pH: float = 7.4,
        **kwargs
    ) -> tuple:
        """
        Creates a voxelized representation of target that contains structural
        information regarding the spatial distribution of its physico-chemical
        properties. It is powered by the Moleculekit library.

        Args:
            target (str): Path to the PDB file or PDB code of the target
                protein.
            voxelsize (float, optional): Distance between the different points
                in the voxelization. Defaults to 1.0.
            buffer (int, optional): Space generated around the protein so that
                the model might analyse its borders. Its value should be as big
                as the outmost layer of the CNN model. Defaults to 6.
            validitychecks (bool, optional): Whether the voxelization program
                will check the integrity of the PDB file. Its best to set it to
                False, specially in the case of screening. Defaults to False.
            channels (list, optional): List of channels allowed for
                voxelization. In any standard use the channels are limited by
                the training of the CNN model and should not be changed.
                However, in case anyone has developed their own model they can
                be changed. Defaults to CHANNELS.

        Raises:
            TypeError: In case the target is neither a PDB code nor a Path to a
                local PDB file.

        Returns:
            tuple: (final_vox, p_centers, nvoxels) First element is the
                voxelized representation of the protein; second element, the
                real cartesian coordinates of each of the voxels in the first
                element; the third element, is the number of such voxels in
                the image.
        """
        if isinstance(target, str):
            protein = Molecule(target, validateElements=False)
        elif isinstance(target, Molecule):
            protein = target
        else:
            raise TypeError(
                'Target input has to be either: a) PDB ID, b) path to local\
                    PDB file, c) Molecule object.'
            )

        protein.remove("not protein")
        protein = prepareProteinForAtomtyping(protein, verbose=0, pH=pH)
        centers = getCenters(protein)
        uchannels = np.ones((len(centers[0]), 8))
        uchannels, undesired_channels = get_undesired_channels(
            uchannels, channels
        )
        vox, p_centers, p_N = getVoxelDescriptors(
            protein,
            voxelsize=voxelsize,
            buffer=buffer,
            validitychecks=validitychecks,
            userchannels=uchannels
        )

        new_vox = select_desired_channels(
            vox, undesired_channels, len(p_centers)
        )

        # From the 2D output create the proper 3D output
        nchannels = new_vox.shape[1]
        final_vox = new_vox.transpose().reshape(
            [1, nchannels, p_N[0], p_N[1], p_N[2]]
        )
        nvoxels = np.array([p_N[0], p_N[1], p_N[2]])

        return final_vox, p_centers, nvoxels

    def cnn(
        self,
        vox: np.array,
        p_centers: np.array,
        stride: int = 1,
        voxelsize: float = 1.0,
        region_size: int = 12,
        occupancy_restrictions: float = 0.4,
        **kwargs
    ) -> np.array:
        border = region_size // 2
        size_x, size_y, size_z = vox.shape[2], vox.shape[3], vox.shape[4]
        max_x, max_y, max_z = size_x-border, size_y-border, size_z-border

        # Create an empty representation of the protein to be filled
        # with the coordination score prediction
        scores = np.zeros((size_x, size_y, size_z))
        scores = torch.Tensor(scores).to('cpu')
        vox = torch.tensor(vox, device=self.device).float()

        x_dim = len(range(border, max_x, stride))
        y_dim = len(range(border, max_y, stride))
        z_dim = len(range(border, max_z, stride))
        num_points = x_dim * y_dim * z_dim
        counter = 0

        for x in tqdm(range(border, max_x, stride)):
            for y in range(border, max_y, stride):
                for z in range(border, max_z, stride):
                    counter += 1

                    if vox[:, 5, x, y, z] > occupancy_restrictions:
                        continue

                    x1, x2 = x - border, x + border
                    y1, y2 = y - border, y + border
                    z1, z2 = z - border, z + border
                    output = self.model(
                        vox[:, :, x1: x2, y1: y2, z1: z2]
                        ).detach().cpu()
                    scores[x, y, z] = output

        print()
        scores = scores.reshape(-1)
        scores = scores.detach().cpu().numpy()
        scores = np.c_[p_centers, scores]

        return scores

    def clusterize(
        self, scores, molecule, cluster_radius
    ):
        clustering = Birch(threshold=cluster_radius, n_clusters=None)
        try:
            labels = clustering.fit_predict(scores[:, :3])
        except ValueError:
            return [], []
        clusters = {}
        mean_clusters = {}
        result = ordered_list()
        for idx, score in enumerate(scores):
            try:
                clusters[labels[idx]].append(score)
            except KeyError:
                clusters[labels[idx]] = []

        for name, cluster in clusters.items():
            cluster = np.array(cluster)
            if len(cluster) == 0:
                continue
            try:
                cluster_mean = np.average(
                    cluster[:, :3], axis=0, weights=cluster[:, 3] ** 3
                )
            except ZeroDivisionError:
                cluster_mean = np.average(cluster[:, :3], axis=0)
            score_sum = np.sum(cluster[:, 3])
            score_mean = np.mean(cluster[:, 3])
            result.add(cluster_mean, score_sum)
            str_mean = ",".join(
                str(round(float(cluster_mean[i]), 3)) for i in range(3)
            )
            mean_clusters[str_mean] = score_mean
        return result, mean_clusters

    def print_clusters(
        self,
        clusters,
        molecule,
        outputfile,
        coordinators,
        cluster_radius,
        args
    ):
        coors = coordinators
        outputfile_name = f'{outputfile}_clusters.txt'
        with open(outputfile_name, 'w') as writer:
            if args is None:
                writer.write('Warning: Configuration Unknown. Check script.')
            else:
                writer.write(f'Configuration: {args}\n')
            writer.write(f'Coordinators: {coordinators}\n')
            title = 'cluster  |  coordinators  |  score\n'
            writer.write(title)
            title = title.strip('\n')
            writer.write('-------     ------------     -----\n')
            for name, center in enumerate(clusters):
                coordinator_found = False
                line = []
                for residue in molecule.residues:
                    for atom in residue.atoms:
                        if atom.element not in ['N', 'O', 'S']:
                            continue
                        if atom.name in ['N', 'O']:
                            message = f'BACK_{atom.name}_'
                        else:
                            message = ''
                        if atom.resname not in coors:
                            continue
                        if (
                            np.linalg.norm(
                                atom - center
                            ) < cluster_radius
                        ):
                            if not coordinator_found:
                                entry = f"{' ' * (3 - len(str(name)))}{name}"
                                entry += "  |  "
                                writer.write(entry)
                            coordinator_found = True
                            res = residue.id.split('_')
                            res2 = res[1].split('(')
                            res2[1] = res2[1].strip(')')
                            message += f"{res[0]}:{res2[0]}:{res2[1]}"
                            if message not in line:
                                line.append(message)
                if coordinator_found:
                    mes = ';'.join(res for res in line)
                    if len(mes) < len(title):
                        mes += ' ' * (len(mes) - len(title))
                    writer.write(';'.join(res for res in line))
                    writer.write(f"  |  {round(clusters.counts(name), 2)}\n")

    def print_PDB(
        self,
        target: str,
        outputfile: str,
        scores: np.array,
        threshold: float,
        cluster_scores: dict,
        molecule: protein,
        **kwargs
    ) -> None:
        """
        Generate a PDB-style file with the coordinates of the probes
        with a score superior to `threshold`. The probes will be
        stored as `HE` atoms and the cluster centers as `AR` atoms. The
        `b-factor` will be used to store the score each probe has obtained.

        Before saving any probe coordinate, it will verify whether the probe
        is at a reasonable distance from any relevant protein atom to mitigate
        possible noise.

        Args:
            target (str): Name of the protein used for the computation.
            outputfile (str): Name of the output file, will be completed with
                the tag `_brigit.pdb` to differenciate it from other output
                files.
            scores (dict): Dict with probe coordinates and their
                coordination scores. Dimensions will be (len(probes), 4).
            threshold (float): Coordination score value below which probes will
                be discarded.
            centers (np.array): Array with cluster center coordinates and their
                coordination scores. Similar to `scores`, its dimensions will
                be (len(cluster_centers), 4).
            molecule (protein): Protein object.
        """
        outputfile = f'{outputfile}_brigit.pdb'
        with open(outputfile, "w") as fo:
            num_at = 0
            num_res = 0
            for entry in scores:

                num_at += 1
                num_res = 1
                ch = "A"
                prb_str = ""

                for idx in range(3):
                    number = str(round(float(entry[idx]), 3))
                    prb_center = "{:.8s}".format(number)
                    if len(prb_center) < 8:
                        prb_center = " "*(8-len(prb_center)) + prb_center
                        prb_str += prb_center
                    else:
                        prb_str += prb_center

                atom = "HE"
                blank = " "*(7-len(str(num_at)))
                fo.write("ATOM" + blank + "%s  %s  SLN %s" %
                         (num_at, atom, ch))
                blank = " "*(3-len(str(num_res)))
                score = str(round(entry[3], 2))
                score = score if len(score) == 4 else score + '0'
                fo.write(blank + "%s     %s  1.00  %s          %s\n" %
                         (num_res, prb_str, score, atom))

            for name, entry in enumerate(cluster_scores.keys()):
                num_at += 1
                num_res = 1
                ch = "A"
                prb_str = ""
                list_entry = entry.split(',')

                for idx in range(3):
                    number = str(round(float(list_entry[idx]), 3))
                    prb_center = "{:.8s}".format(number)
                    if len(prb_center) < 8:
                        prb_center = " "*(8-len(prb_center)) + prb_center
                        prb_str += prb_center
                    else:
                        prb_str += prb_center

                atom = "AR"
                blank = " "*(7-len(str(num_at)))
                fo.write("ATOM" + blank + "%s  %s  SLN %s" %
                         (num_at, atom, ch))
                blank = " "*(3-len(str(num_res)))
                score = str(cluster_scores[entry])
                score = score if len(score) == 4 else score + '0'
                fo.write(blank + "%s     %s  1.00  %s          %s\n" %
                         (num_res, prb_str, score, atom))

    def filter_best_results(
        self,
        coor_scores: np.array,
        combined_threshold: float,
    ) -> np.array:
        # Selection of best positions
        best_scores = np.argwhere(coor_scores[:, 3] > combined_threshold)
        new_scores = np.zeros((len(best_scores), 4))
        for idx, idx_score in enumerate(best_scores):
            new_scores[idx, :] = coor_scores[idx_score, :]
        return new_scores


def analyse_probes(
    molecule, probes, stats, gaussian_stats, molecule_info,
    coordinators, residue_score, max_coordinators,
    threshold, cnn_weight
):
    zeros = np.zeros((len(probes), 4))
    coor_weight = (1 - cnn_weight)

    for idx, probe in enumerate(probes):
        if probe[3] > threshold:
            zeros[idx, :] = probe
            zeros[idx, 3] *= cnn_weight
            zeros[idx, 3] += coordination_scorer(
                molecule=molecule,
                probe=probe,
                stats=stats,
                gaussian_stats=gaussian_stats,
                molecule_info=molecule_info,
                coordinators=coordinators,
                residue_scoring=gaussian_score,
                max_coordinators=max_coordinators
            ) * coor_weight
    return zeros


def run(args: dict):
    print(args, end="\n\n")
    args['args'] = args
    if args['threads'] == 0:
        args['threads'] = multiprocessing.cpu_count() * 2
    brigit = Brigit(**args)
    brigit.predict(**args)


if __name__ == '__main__':
    help(Brigit)



"""
PDB Parser utilities within the DeepBioMetAll tool.

Contains 3 classes:
    - protein
    - residue
    - atom

Contains 1 function:
    - write_site

Simplifies the process of parsing PDB files and
solves some of the problems with BioPython.

Copyrigth by Raúl Fernández Díaz
"""
import os
import numpy as np
from .data import METAL_RESNAMES, NEW_METALS
from .tools import download_pdb


def res2letter(residue: str):
    global RESNAME2LETTER_DICT
    try:
        result = RESNAME2LETTER_DICT[residue]
    except KeyError:
        result = None
    return result


class atom():
    """
    Atom class that stores all significant information from a
    protein atom.

    Attributes:
        entry_type (str): ATOM or HETATM.
        id (int): Atom serial number.
        name (str): Role that the atom plays in the protein.
        element (str): IUPAQ symbol of the atom.
        resname (str): Name of the residue to which it
            belongs.
        res_ID (str): Has been adapted to accomodate the MetalPDB format.
            Example: LYS_457(B).
        coordinates (np.array): 3D coordinates of the atom.
    """

    def __init__(self, line: str):
        """
        Instanciate new atom class.

        Args:
            line (str): PDB line describing the atom
        """
        self.entry_type = 'ATOM' if line[:4] == 'ATOM' else 'HETATM'
        self.id = int(line[6:11])
        self.name = _remove_blank_spaces(line[12:16])
        self.element = _remove_blank_spaces(line[76:80])
        self.resname = _remove_blank_spaces(line[17:20])
        self.res_ID = f'{self.resname}_{int(line[22:26])}({line[21]})'
        self.coordinates = np.array(
            [[float(line[30:38]), float(line[38:46]), float(line[48:54])]]
        )
        self.companions = [self.element]

    def __str__(self):
        return f'{self.id}_{self.name}'

    def __repr__(self):
        return f'{self.id}_{self.name}'

    def __sub__(self, other):
        if isinstance(other, atom):
            return self.coordinates - other.coordinates
        else:
            return self.coordinates - other


class residue():
    """
    Residue class that stores all significant information from a protein
    residue.
    """

    def __init__(self, atoms: list):
        self.name = atoms[0].resname
        self.id = atoms[0].res_ID
        self.atoms = atoms
        alpha, beta, o, n = self.coordinates()
        self.alpha = alpha
        self.beta = beta
        self.o = o
        self.n = n

    def coordinates(self):
        """
        Calculates the center of mass of the residue.

        Uses a weighted average function from the numpy library. Results
        are then rounded as per the significant numbers. Takes advantage of a
        global variable, i.e., `ATOMIC_MASSES` that contains the atomic masses
        of the main atoms that might comprise in the residue.
        """
        global ATOMIC_MASSES
        alpha = None
        beta = None
        o = None
        n = None

        for atom in self.atoms:
            if atom.name == 'CA':
                alpha = atom
            elif (atom.name == 'CB'):
                beta = atom
            elif atom.name == 'O':
                o = atom
            elif atom.name == 'N':
                n = atom

        return alpha, beta, o, n

    def __str__(self):
        return f'{self.name}'

    def __repr__(self):
        return f'{self.name}'

    def __sub__(self, other):
        return self.coordinates - other.coordinates


class protein():
    """
    Protein class that stores all relevant information from
    a PDB file.

    Attributes:
        name (str): PDB ID.
        atoms (list): List of atoms comprising the protein.
        hetatms (list): List of hetatms comprising the
            protein.

    Methods:
        write_MFS (path, center, radius): Writes a PDB
            file with the region surrounding a given
            HETATM.
    """

    def __init__(self, pdb_code: str, residues: bool = False):
        """
        Instanciates a new protein class.

        Parses a PDB file and collects all its atoms
        and hetatms for further processing. If the PDB
        file is not available locally, it downloads it from
        the appropriate server.

        Args:
            pdb_code (str): Path to a PDB file.
        """
        self.name = pdb_code
        self.atoms = []
        self.hetatms = []
        self.residues = []

        self._calculate_atoms(pdb_code)
        self._calculate_residues(residues)

    def _calculate_residues(self, residues):
        if not residues:
            return None

        try:
            residue_ids = {self.atoms[0].res_ID}
        except IndexError:
            self.residues = None
            return None

        tmp_id = self.atoms[0].res_ID
        tmp_atoms = []

        for atom in self.atoms:
            if (atom.res_ID not in residue_ids):
                tmp_id = atom.res_ID
                residue_ids.add(atom.res_ID)
                self.residues.append(residue(tmp_atoms))
                tmp_atoms = [atom]

            elif atom.res_ID == tmp_id:
                tmp_atoms.append(atom)

    def _calculate_atoms(self, pdb_code):
        try:
            self._read_pdb(pdb_code)

        except FileNotFoundError:
            filename = download_pdb(pdb_code, '.')
            self._read_pdb(filename)
            # os.remove(filename)
        except IsADirectoryError:
            filename = download_pdb(pdb_code, '.')
            self._read_pdb(filename)
            # os.remove(filename)

    def _read_pdb(self, pdb_code):
        with open(pdb_code) as pdbfile:
            for line in pdbfile:
                if line[:4] == 'ATOM':
                    self.atoms.append(atom(line))
                elif line[:6] == 'HETATM':
                    self.hetatms.append(atom(line))

    def find_metals(self):
        to_remove = []
        metals = []
        self.metals = []

        for hetatm in self.hetatms:
            if (
                hetatm.element in NEW_METALS and
                hetatm.resname in METAL_RESNAMES
            ):
                metals.append(hetatm)

        for idx, metal in enumerate(metals):
            if idx in to_remove:
                continue
            distances = np.ones(len(metals))
            distances[:] = 5.0
            for idx2, metal2 in enumerate(metals):
                if idx2 in to_remove or idx == idx2:
                    continue
                if np.linalg.norm(metal - metal2) < 5.0:
                    metal.companions.append(metal2.element)
                    to_remove.append(idx2)
            self.metals.append(metal)

    def __str__(self):
        return f'Protein Structure of PDB ID: {self.name}'

    def __repr__(self):
        return str(self)


def write_site(
    structure: protein,
    output_path: str,
    center: np.array,
    radius: float,
    protein_atoms: bool = True
):
    """
    Generate a PDB with a all residues that contain at least one
    atom within a certain radius of a set of coordinates.

    Args:
        output_path (str): Path where the new PDB
            file will be written.
        center (np.array): Coordinates for the center of the site with
            shape (1, 3)
        radius (float): Radius of the site to be
            computed.
    """
    mfs_residues = set()

    for atom in structure.atoms:
        if np.linalg.norm(atom - center) <= radius:
            mfs_residues.add(atom.res_ID)

    control_num_atoms = []
    atoms_in_trp = 27
    threshold = (1/3) * (radius ** 2) * atoms_in_trp * np.pi

    with open(output_path, 'w') as fo:
        for atom in structure.atoms:
            if atom.res_ID in mfs_residues:
                fo.write(atom.line)
                control_num_atoms.append(atom)
        fo.write('END')

    if protein_atoms:
        if len(control_num_atoms) > threshold:
            try:
                os.remove(output_path)
            except FileNotFoundError:
                pass
            raise RuntimeError(f"Too Many Atoms: {len(control_num_atoms)}")

        elif len(control_num_atoms) == 0:
            try:
                os.remove(output_path)
            except FileNotFoundError:
                pass
            raise RuntimeError(f"Not Enough Atoms: {len(control_num_atoms)}")


def _remove_blank_spaces(sentence):
    new_sentence = ""
    for char in sentence:
        if char not in [" ", "\t", "\n"]:
            new_sentence += char
    return new_sentence


if __name__ == '__main__':
    help(protein)
    help(residue)
    help(atom)
    help(write_site)



"""
Layers module within the Brigit package.

Contains the architecture of layers used for building the models.

Copyright by Raúl Fernández Díaz
"""
import torch.nn as nn


class conv3D (nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: int = 0,
        activation: nn.Module = nn.ReLU,
        dropout: float = 0.0,
        stride: int = 1,
    ):
        super().__init__()
        self.conv = nn.Conv3d(
            in_channels, out_channels, kernel_size, stride, padding
            )
        self.activation = activation()
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.activation(self.conv(x)))


class linear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        dropout: float = 0.0,
        activation: nn.Module = nn.ReLU
    ):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features)
        self.activation = activation()
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.activation(self.linear(x)))


if __name__ == '__main__':
    help(conv3D)
    help(linear)



"""
Execution BioBrigit module.

Copyrigth by Raúl Fernández Díaz
"""
import argparse
import multiprocessing

import biobrigit


def parse_cli() -> dict:
    """
    Console script for brigit.
    """
    p = argparse.ArgumentParser()
    p.add_argument("target", type=str,
                   help="Molecule PDB file to be analysed.")
    p.add_argument("metals", type=str,
                   help="Symbol of the metal that is to be located. More than one metal can be used by separting them with a comma: METAL1,METAL2")
    p.add_argument("--device", type=str, default='cuda',
                   help="Device in which calculations will be run.")
    p.add_argument("--device_id", type=int, default=0,
                   help="GPU ID in which the calculations will be run.")
    p.add_argument("--outputfile", type=str, default=None,
                   help='Path where the output should be written.')
    p.add_argument("--max_coordinators", type=int, default=2,
                   help="Number of residues that need to be coordinating a\
                   given metal.")
    p.add_argument("--residues", type=int, default=10,
                   help="Number of most likely residues to consider.")
    p.add_argument(
        "--stride", type=int, default=1,
        help="Step of the sliding window when evaluating the protein."
    )
    p.add_argument("--pH", type=float, default=7.4,
                   help='pH where the structure is to be evaluated.')
    p.add_argument("--cluster_radius", type=float, default=5.0,
                   help="Threshold used for the Birch clustering algorithm.")
    p.add_argument("--cnn_threshold", type=float, default=0.5,
                   help="Threshold for considering CNN points as possible\
                    coordinations.")
    p.add_argument("--combined_threshold", type=float, default=0.5,
                   help="Threshold for considering predictions positive\
                    combining BioMetAll and CNN scores.")
    p.add_argument("--voxelsize", type=float, default=1.0,
                   help="Resolution of the 3D representation. In Arnstrongs.")
    p.add_argument("--cnn_weight", type=float, default=0.5,
                   help='Weight assigned to CNN.')
    p.add_argument(
        "--verbose", type=int, default=1,
        help="Information that will be displayed. 0: Only Moleculekit, 1: All."
    )
    p.add_argument("--residue_score", type=str, default='gaussian',
                   help='Scoring function for residue coordination analysis.')
    p.add_argument("--threads", type=int, default=0,
                   help='Number of threads available for multithreading\
                    calculation. By default it will create 2 threads per\
                    physical core.')
    args = vars(p.parse_args())
    args['metals'] = args['metals'].split(',')
    print()
    return args


def welcome() -> None:
    """
    Print a welcoming message when the program is executed.
    """
    message = "Using BioBrigit by Raúl Fernández-Díaz"
    print("-" * (len(message) + 4))
    print("| " + message + " |")
    print("-" * (len(message) + 4))
    print()


def main():
    """
    Execute the program.
    """
    multiprocessing.freeze_support()
    welcome()
    args = parse_cli()
    biobrigit.run(args)


if __name__ == '__main__':
    main()



"""
Various utility functions within the Brigit package.

Contains 9 functions:
    - download_pdb
    - get_undesired_channels
    - select_desired_channels
    - voxelize
    - distribute
    - find_coordinators
    - geometry
    - set_up_cuda
    - load_model

1 class:
    - ordered_list

Includes functions and data sets that are required in different
files throughout the project, thus, preventing redundancies.

Copyright by Raúl Fernández Díaz
"""

import os
import sys
import urllib.request

import numpy as np
import torch
from moleculekit.molecule import Molecule
from moleculekit.tools.atomtyper import prepareProteinForAtomtyping
from moleculekit.tools.voxeldescriptors import getCenters, getVoxelDescriptors

from .data import CHANNELS_DICT, read_stats
from .models import BaseModel, BrigitCNN


class ordered_list():
    def __init__(self):
        self.list = []
        self.dict = {}

    def add(self, other, counts):
        for idx, item in enumerate(self.list):
            if self.counts(idx) < counts and counts > 0.0:
                self.list.insert(idx, other)
                self.dict[str(other)] = counts
                return

        self.list.append(other)
        self.dict[str(other)] = counts

    def pop(self, idx):
        del self.dict[str(self.list[idx])]
        self.list.pop(idx)

    def counts(self, idx):
        return self.dict[str(self.list[idx])]

    def to_list(self):
        return self.list

    def __len__(self):
        return len(self.list)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        return str(self.list)

    def __iter__(self):
        self.n = 0
        self.max = len(self.list)
        return self

    def __next__(self):
        if self.n < self.max:
            result = self.list[self.n]
            self.n += 1
        else:
            raise StopIteration

        return result


def download_pdb(pdbcode: str, datadir: str,
                 downloadurl: str = 'https://files.rcsb.org/download/',
                 verbose: bool = True):
    """
    Downloads a PDB file from the RCSB-PDB and saves it in a directory.

    Args:
        pdbcode (str): The standard PDB ID e.g. '3ICB' or '3icb'.
        datadir (str): The directory where the downloaded file will be saved.
        downloadurl (str, optional): The base PDB download URL. Defaults to
        'https://files.rcsb.org/download/'.

    Returns:
        output_filename (str): Path of the PDB file.
    """
    pdb_filename = pdbcode + ".pdb"
    url = downloadurl + pdb_filename
    output_filename = os.path.join(datadir, pdb_filename)

    try:
        urllib.request.urlretrieve(url, output_filename)
    except Exception as err:
        if verbose:
            print(str(err), file=sys.stderr)
        return None

    return output_filename


def get_undesired_channels(uchannels, channels):
    """
    Helper function to voxelize() that reads the desired channels for
    voxelization and determines what channels should be turned off.
    It outputs a list of ints with the indexes of the undesired channels.
    """
    global CHANNELS
    working_channels = []
    undesired_channels = []

    for channel in channels:
        if channel not in CHANNELS_DICT.keys():
            raise Exception(
                f"Unsupported channel: {channel}"
                + f"Supported channels: {CHANNELS_DICT.keys()}."
                )
        working_channels.append(CHANNELS_DICT[channel])

    for i in range(8):
        if i not in working_channels:
            undesired_channels.append(i)
    for channel in undesired_channels:
        uchannels[:, channel] = 0

    return uchannels, undesired_channels


def select_desired_channels(protein_vox: np.array, undesired_channels: list,
                            length_centers: int) -> np.array:
    """
    From the moleculekit protein_vox extract only the desired channels.

    Args:
        protein_vox (np.array): Protein voxelized representation.
        undesired_channels (list): List of channels to remove.
        length_centers (int): Number of voxels in the representation.

    Returns:
        np.array: Protein voxelized representation without the undesired
                channels.
    """
    new_protein_vox = np.zeros(
        (length_centers, 8-len(undesired_channels))
    )
    j = 0

    for i in range(8):
        if i not in undesired_channels:
            new_protein_vox[:, j] = protein_vox[:, i]
            j += 1

    return new_protein_vox


def distribute(list, threads):
    length = len(list)
    sublist = [[] for idx in range(threads)]
    i = 0
    j = 0
    while i < length:
        if j == threads:
            j = 0
        sublist[j].append(list[i])
        j += 1
        i += 1
    return sublist


def voxelize(
    path: str,
    site_center: list,
    channels: list,
    buffer: int,
    voxel_resolution: float,
    radius: int,
    protein_atoms: bool = True
):
    """
    Generate voxel representation of a given MFS.

    Args:
        path (str): Path where the PDB file storing the MFS.
        metal_center (list): Coordinates of the metal center.
        channels (list): Channels to be computed during voxelization.
        buffer (int): Number of null voxels to add in every dimension.
        voxel_resolution (float): Resolution of the voxelization. Units in
                                    Arngstrons.
        radius (float): Dimension of the boxes.

    Returns:
        voxel_MFS (torch.Tensor): Voxel representation of the MFS as a Tensor.
    """

    protein = Molecule(path, validateElements=False)
    protein.remove('not protein')

    if protein.numAtoms < 1 and protein_atoms:
        raise RuntimeError('There are no protein atoms involved in the MFS.')

    protein = prepareProteinForAtomtyping(protein, verbose=False)

    centers = getCenters(protein)
    uchannels = np.ones((len(centers[0]), 8))
    uchannels, undesired_channels = get_undesired_channels(uchannels, channels)

    voxels, centers, dimns = getVoxelDescriptors(
                                                protein,
                                                voxelsize=voxel_resolution,
                                                buffer=buffer,
                                                validitychecks=False,
                                                userchannels=uchannels
                                                )

    new_voxels = select_desired_channels(
        voxels, undesired_channels, len(centers)
    )

    if len(new_voxels) == 0 or np.sum(new_voxels) == 0:
        raise RuntimeError("New voxels = 0")

    # From the 2D output create the proper 3D output
    nchannels = new_voxels.shape[1]
    voxels_t = new_voxels.transpose().reshape(
        [1, nchannels, dimns[0], dimns[1], dimns[2]]
        )
    voxels_t = np.array(voxels_t, dtype='float32')
    center_voxel = [i//2 for i in voxels_t.shape]

    # Ensuring radius is correct type
    radius = int(radius)
    voxelized = torch.empty(
        (1, len(channels), int(radius)*2, int(radius)*2, int(radius)*2)
    )
    voxelized[:, :, :, :, :] = torch.from_numpy(
        voxels_t[:, :, center_voxel[2]-radius:center_voxel[2]+radius,
                 center_voxel[3]-radius:center_voxel[3]+radius,
                 center_voxel[4]-radius:center_voxel[4]+radius]
        )
    return voxelized


def find_coordinators(metal: str, num_residues: int = 20):
    stats, gaussian_stats = read_stats()
    stats = stats[metal]
    gaussian_stats = gaussian_stats[metal]
    residues = ordered_list()

    for residue, res_stats in stats.items():
        if residue not in gaussian_stats.keys():
            continue
        if len(residues) < num_residues:
            if res_stats['fitness'] > 0.0:
                residues.add(residue, res_stats['fitness'])
        elif res_stats['fitness'] > residues.counts(-1):
            residues.pop(-1)
            residues.add(residue, res_stats['fitness'])

    coordinators = residues.to_list()

    return coordinators, stats, gaussian_stats


def load_stats(metal: str, residues: dict, backbone: bool):
    stats, gaussian_stats = read_stats()
    stats = stats[metal]
    gaussian_stats = gaussian_stats[metal]
    if backbone:
        coordinators = {
            'residue': residues,
            'backbone_o': residues,
            'backbone_n': residues
        }
    else:
        coordinators = {
            'residue': residues,
            'backbone_o': {'mandatory': {}, 'either': [[]]},
            'backbone_n': {'mandatory': {}, 'either': [[]]}

        }

    return coordinators, stats, gaussian_stats


def geometry(v1, v2):
    v1_distances = np.linalg.norm(v1, axis=1)
    v2_distances = np.linalg.norm(v2, axis=1)
    v1_v2_distances = np.linalg.norm(v1 - v2, axis=1)
    if np.any(v1_v2_distances[:] == 0.0):
        return None, None, None
    v1v2_angles = np.arccos(
        (
            np.square(v1_distances) + np.square(v1_v2_distances) -
            np.square(v2_distances)
        ) /
        (2 * v1_distances * v1_v2_distances)
    )
    return v1_distances, v2_distances, v1v2_angles


def set_up_cuda(device_id: int) -> None:
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.fastest = True
    torch.cuda.set_device(device_id)


def load_model(device: str, **kwargs) -> BaseModel:
    path = os.path.join(
        os.path.dirname(__file__), "trained_models",
        f'BrigitCNN.bak'
    )
    # if model.lower() == 'brigitcnn':
    model = BrigitCNN.load_from_checkpoint(
        path,
        map_location=device,
        learning_rate=2e-4,
        neurons_layer=64,
        size=12,
        num_dimns=6
    )

    # elif model.lower() == 'tinybrigit':
    #     print(torch.load(path)['state_dict'].keys())
    #     model = BrigitCNN.load_from_checkpoint(
    #         path,
    #         map_location=device,
    #         learning_rate=2e-4,
    #         neurons_layer=32,
    #         size=12,
    #         num_dimns=6
    #     )
    #     # val = torch.load(path, map_location='cpu')
    # else:
    #     message = f'Model: {model} is not available.'
    #     message += ' Please select BrigitCNN.'
    #     raise RuntimeError(message)
    model.to(device)
    model.eval()
    return model


if __name__ == '__main__':
    help(download_pdb)
    help(get_undesired_channels)
    help(select_desired_channels)
    help(voxelize)
    help(geometry)
    help(set_up_cuda)



"""
Models module within the Brigit package.

Contains the architecture of the models used
for the project.

Copyright by Raúl Fernández Díaz
"""

import torch
import numpy as np
import sklearn.metrics as skmetrics
from torch import nn
import pytorch_lightning as pl
import torchmetrics.functional as metrics
from .layers import conv3D, linear


class BaseModel(pl.LightningModule):
    def __init__(self):
        super().__init__()
        self.save_hyperparameters()
        self.loss = nn.BCELoss()

    def forward(self, x):
        return self.layers(x)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate
        )
        return optimizer

    def training_step(self, train_batch, batch_idx):
        images, labels = train_batch
        predictions = self(images)
        loss = self.loss(predictions, labels.float())

        return {
            'labels': labels,
            'predictions': predictions,
            'loss': loss
        }

    def validation_step(self, batch, batch_idx):
        images, labels = batch
        predictions = self(images)
        loss = self.loss(predictions, labels.float())

        return {
            'val_labels': labels,
            'val_predictions': predictions,
            'val_loss': loss
        }

    def training_epoch_end(self, outputs):
        labels = torch.cat([x['labels'] for x in outputs])
        predictions = torch.cat([x['predictions'] for x in outputs])
        loss = self.loss(predictions, labels.float())
        labels = labels.int()
        acc = metrics.accuracy(predictions, labels)

        self.log("train_acc", acc)
        self.log("train_loss", loss)

    def validation_epoch_end(self, outputs):
        labels = torch.cat([x['val_labels'] for x in outputs])
        predictions = torch.cat([x['val_predictions'] for x in outputs])
        loss = self.loss(predictions, labels.float())
        labels = labels.int()
        acc = metrics.accuracy(predictions, labels)
        prec, rec = metrics.precision_recall(predictions, labels,
                                             average='micro')
        f1 = metrics.f1_score(predictions, labels, average='micro')
        spe = metrics.specificity(predictions, labels, average='micro')
        auroc = metrics.auroc(predictions, labels, num_classes=1,
                              average='micro')

        labels = labels.detach().cpu().numpy()
        predictions = (predictions > 0.5).int().detach().cpu().numpy()
        cohen = torch.from_numpy(
            np.array(skmetrics.cohen_kappa_score(labels, predictions))
        )
        mcc = torch.from_numpy(
            np.array(skmetrics.matthews_corrcoef(labels, predictions))
        )
        log_loss = torch.from_numpy(
            np.array(skmetrics.log_loss(labels, predictions))
        )

        self.log("acc", acc)
        self.log("loss", loss)
        self.log("f1", f1)
        self.log("auroc", auroc)
        self.log("precision", prec)
        self.log("recall", rec)
        self.log("specificity", spe)
        self.log("cohen", cohen)
        self.log("mcc", mcc)
        self.log("log_loss", log_loss)


class BrigitCNN(BaseModel):
    def __init__(
        self,
        learning_rate: float,
        neurons_layer: int = 32,
        size: int = 12,
        num_dimns: int = 6,
        **kwargs
    ):
        super().__init__()

        self.save_hyperparameters(
            "learning_rate",
            "neurons_layer",
            "num_dimns",
        )
        self.learning_rate = learning_rate
        self.example_input_array = torch.rand(
            num_dimns*size**3, 1).view(1, num_dimns, size, size, size)

        self.convolutional = nn.Sequential(
            nn.Dropout3d(0.2),
            conv3D(num_dimns, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            nn.BatchNorm3d(neurons_layer),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            nn.BatchNorm3d(neurons_layer),
            nn.AvgPool3d(2),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            nn.BatchNorm3d(neurons_layer),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            conv3D(neurons_layer, neurons_layer, 3, 1, nn.LeakyReLU, 0.2),
            nn.BatchNorm3d(neurons_layer),
            nn.AvgPool3d(2),
            nn.Flatten()
        )
        self.classifier1 = nn.Sequential(
            linear(neurons_layer*3**3, 512, 0.2, nn.LeakyReLU),
            linear(512, 256, 0.2, nn.LeakyReLU)
        )
        self.classifier2 = nn.Sequential(
            linear(256, 1, 0.0, nn.Sigmoid)
        )

    def forward(self, x):
        x = self.convolutional(x)
        x = self.classifier1(x)
        return self.classifier2(x)


if __name__ == '__main__':
    help(BaseModel)
    help(BrigitCNN)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
readme = (this_directory / "README.md").read_text()

requirements = [
    'scikit-learn',
    'torch',
    'moleculekit',
    'lightning',
    'pdb2pqr'
]

test_requirements = requirements

setup(
    author="Raúl Fernández Díaz",
    author_email='raul.fernandezdiaz@ucdconnect.ie',
    python_requires='>=3.6',
    classifiers=[
    ],
    description="Computational tool for the prediction of metal-binding loading paths in proteins using deep convolutional neural networks.",
    entry_points={
        'console_scripts': [
            'biobrigit=biobrigit.__main__:main',
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    # package_data={'': ['*.ckpt', '*.json']},
    data_files=[('',
                 ['biobrigit/utils/stats/gaussian_statistics.json',
                  'biobrigit/utils/stats/residue_statistics.json',
                  'biobrigit/utils/trained_models/BrigitCNN.bak',
                  'README.md'])],
    keywords='biobrigit',
    name='biobrigit',
    packages=find_packages(),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/insillichem/BioBrigit',
    version='0.0.9',
    zip_safe=False,
)



"""
Brigit module that contains the main functions
for evaluating coordination probes.

Contains 5 functions:

    - parse_residues
    - coordination_scorer
    - discrete_score
    - gaussian_score
    - double_gaussian

Copyright by Raúl Fernández Díaz
"""

import numpy as np

from .pdb_parser import protein
from .tools import geometry


def parse_residues(
    molecule: protein,
    coordinators: dict,
) -> dict:
    """
    Create a dictionary that collects the coordinates of the
    atoms of a given protein that belong to the most likely
    coordinators. Depending on the coordination configuration
    (i.e., residue, backbone O, or backbone N), the atoms of interest
    will vary (e.g., for residue atoms of interest are alpha C and beta
    C).

    Args:
        molecule (protein): Protein object.
        coordinators (dict): Dictionary summarising the residues that
            are the most likely coordinators for a given metal for
            the different configurations.

    Returns:
        info (dict): Collection of the atoms of interest for evaluating
            coordination, grouped by the coordination configuration.
    """
    alphas = {residue: [] for residue in coordinators}
    betas = {residue: [] for residue in coordinators}

    for idx, residue in enumerate(molecule.residues):
        try:
            if residue.name in coordinators:
                alphas[residue.name].append(residue.alpha.coordinates)
                betas[residue.name].append(residue.beta.coordinates)
        except AttributeError:
            continue

    alphas = {
        residue: np.array(alphas[residue]).reshape(len(alphas[residue]), 3)
        for residue in coordinators
    }
    betas = {
        residue: np.array(betas[residue]).reshape(len(betas[residue]), 3)
        for residue in coordinators
    }

    info = {
        'residue_alphas': alphas,
        'residue_2nd_atom': betas,
    }
    return info


def coordination_scorer(
    molecule: protein,
    probe: np.array,
    stats: dict,
    gaussian_stats: dict,
    molecule_info: dict,
    coordinators: dict,
    residue_scoring,
    max_coordinators,
    **kwargs
) -> float:
    """
    Compute how strongly a hypothetical metal would be coordinated
    if it where located at the coordinates of `probe`.

    It can use different scoring functions for residue or backbone
    coordinations. It iterates through all coordinating residues and
    calculates the distances between the probe and the alpha C and a
    second atom (beta C for residue coordination, O or N for the corresponding
    backbone coordination), and the angles between both atoms with regards
    to the probe. Then it uses this information and the appropriate scoring
    function to assign a coordination score to the probe.

    Finally, the cummulative score is multiplied by a factor that depends on
    how many coordinating residues were located and how many were expected.
    Final value will be at maximum of 1.0.

    Args:
        molecule (protein): Protein object.
        probe (np.array): 3D array with the spatial coordinates of a
            hypothetical metal.
        stats (dict): Dictionary containing statistical values describing
            the distances a metallic ion has to keep with regards to different
            backbone atoms in order to properly be coordinated.
        gaussian_stats (dict): Similar to `stats`, but these statistics have
            been calculated to fit the combination of 2 gaussian curves.
        molecule_info (dict): Dictionary containing a collection of the atoms
            of interest for evaluating coordination, grouped by the
            coordination configuration. Output of `parse_residues` function.
        coordinators (dict): Dictionary summarising the residues that
            are the most likely coordinators for a given metal for
            the different configurations.
        residue_scoring (function): Function indicating how to score the
            coordination through residues.
        backbone_scoring (function): Function indicating how to score the
            coordination through backbone atoms.

    Returns:
        score (float): Score associated to the point in space occupied by the
            `probe`.
    """
    fitness = 0
    possible_coordinators = 0

    kwargs['gaussian_stats'] = gaussian_stats
    for res in coordinators:
        alphas = probe[:3] - molecule_info['residue_alphas'][res]
        atm2 = probe[:3] - molecule_info['residue_2nd_atom'][res]
        dist_1, dist_2, angles = geometry(alphas, atm2)
        fitness_res, coors_res = (
            residue_scoring(
                dist_1, dist_2, angles, res, stats, **kwargs
            )
        )
        fitness += fitness_res
        possible_coordinators += coors_res

    fitness *= (possible_coordinators / max_coordinators)
    return fitness if fitness < 1.0 else 1.0


def discrete_score(
    dist_1: np.array,
    dist_2: np.array,
    angles: np.array,
    residue: str,
    stats: dict,
    coor_type: str,
    **kwargs
) -> tuple:
    """
    Scoring function to be used by the `coordination_scorer` function
    to compute the relative strength or suitability of the spatial position
    represented by the probe for coordinating a given metal ion.

    This particular scoring function uses a set of statistical values
    calculated as the median of the distribution of distances between the
    metallic moietie, the alpha C and the appropriate second atom, and the
    angles between them. The ranges are set at the median +- 3 times the
    standard deviation.

    The algorithm first finds all positions that comply with all 3 requirements
    simultaneously and scores them according to the relative abundance of the
    specific metal being coordinating by the specific residue.

    Args:
        dist_1 (np.array): Distances from `probe` to the alpha C.
        dist_2 (np.array): Distances from `probe` to the second atom which
            could be beta C (residue), backbone O, or backbone N.
        angles (np.array): Angles formed by the vectors defined by alpha C and
            `probe` and 2nd atom and `probe`.
        residue (str): Name of the residue currently being evaluated.
        stats (dict):  Dictionary containing statistical values describing
            the distances a metallic ion has to keep with regards to different
            backbone atoms in order to properly be coordinated.
        coor_type (str): Whether the coordination is with the residue, with a
            backbone O, or a backbone N. The main difference is the set of
            statistics used for the evaluation.

    Returns:
        tuple: Contains two different values, i.e., `fitness` which is the
            score obtained for a given probe and a given type of residue and
            `possible_coordinators` which is the number of such residues that
            fulfill the geometric requirements.
    """
    if coor_type == 'residue':
        a, b, c, d, e, f = 'amin', 'amax', 'bmin', 'bmax', 'abmin', 'abmax'
        g = 'fitness'
    elif coor_type == 'backbone_o':
        a, b, c, d, e, f = 'aomin', 'aomax', 'omin', 'omax', 'maomin', 'maomax'
        g = 'o_fitness'
    elif coor_type == 'backbone_n':
        a, b, c, d, e, f = 'anmin', 'anmax', 'nmin', 'nmax', 'manmin', 'manmax'
        g = 'n_fitness'

    fitness = 0
    possible_coordinators = 0
    alpha_trues = np.argwhere(
        (dist_1 > stats[residue][a]) &
        (dist_1 < stats[residue][b])
    )
    beta_trues = np.argwhere(
        (dist_2 > stats[residue][c]) &
        (dist_2 < stats[residue][d])
    )
    ab_trues = np.argwhere(
        (angles > stats[residue][e]) &
        (angles < stats[residue][f])
    )
    for true in alpha_trues:
        if true in beta_trues and true in ab_trues:
            fitness += stats[residue][g]
            possible_coordinators += 1

    return fitness, possible_coordinators


def gaussian_score(
    dist_1: np.array,
    dist_2: np.array,
    angles: np.array,
    residue: str,
    stats: dict,
    gaussian_stats: dict,
    **kwargs
) -> tuple:
    """
    Scoring function to be used by the `coordination_scorer` function
    to compute the relative strength or suitability of the spatial position
    represented by the probe for coordinating a given metal ion.

    This particular scoring function uses a set of statistical values
    calculated as the parameters of the combination of 2 gaussian curves
    that best fit the distribution of the data. Currently, this function
    can only be used for evaluating residue coordination and not backbone
    O or N.

    The algorithm first finds all positions that comply with all 3 requirements
    simultaneously and scores them according to the relative abundance of the
    specific metal being coordinating by the specific residue and to their
    position within the gaussian curves.


    Args:
        dist_1 (np.array): Distances from `probe` to the alpha C.
        dist_2 (np.array): Distances from `probe` to the second atom which
            could be beta C (residue), backbone O, or backbone N.
        angles (np.array): Angles formed by the vectors defined by alpha C and
            `probe` and 2nd atom and `probe`.
        residue (str): Name of the residue currently being evaluated.
        stats (dict):  Dictionary containing statistical values describing
            the distances a metallic ion has to keep with regards to different
            backbone atoms in order to properly be coordinated.
        gaussian_stats (dict):  Similar to `stats`, but these statistics have
            been calculated to fit the combination of 2 gaussian curves.

    Returns:
        tuple: _description_
    """
    fitness = 0
    possible_coordinators = 0

    score_1 = bimodal(dist_1, *gaussian_stats[residue]['alpha'])
    score_2 = bimodal(dist_2, *gaussian_stats[residue]['beta'])
    score_3 = bimodal(angles, *gaussian_stats[residue]['MAB'])

    # BioMetAll v1.0 filter
    alpha_trues = np.argwhere(score_1 > 0.01)
    beta_trues = np.argwhere(score_2 > 0.01)
    ab_trues = np.argwhere(score_3 > 0.01)
    true_indexes = []

    for true in alpha_trues:
        if true in beta_trues and true in ab_trues:
            true_indexes.append(true)

    if len(true_indexes) < 1:
        return 0, 0

    # Gaussian score
    true_indexes = np.array(true_indexes)

    fitness = np.sum(
        (score_1[true_indexes] + score_2[true_indexes] + score_3[true_indexes])
        / 3
    )
    fitness = float(fitness)
    fitness *= stats[residue]['fitness']
    possible_coordinators += len(true_indexes)

    return fitness, possible_coordinators


def bimodal(
    x: np.array,
    chi1: float,
    nu1: float,
    sigma1: float,
    chi2: float,
    nu2: float,
    sigma2: float
) -> np.array or float:
    """
    Helper function for the `gaussian_score` function that computes
    the score associated to a certain set of parameters for the input
    `x`.

    Args:
        x (np.array): Set of input values to evaluate.
        prop (float): Factor describing the contribution of each of
            the gaussians to the final result.
        chi1 (float): Height of the first gaussian.
        nu1 (float): Average value for the first gaussian.
        sigma1 (float): Standard deviation of the first gaussian.
        chi2 (float): Height of the second gaussian.
        nu2 (float): Average value for the second gaussian.
        sigma2 (float): Standard deviation of the second gaussian.

    Returns:
        result (np.array or float): Value or array of values with len(x)
            with the associated probability.
    """
    first_gaussian = _normpdf(x, chi1, nu1, sigma1)
    second_gaussian = _normpdf(x, chi2, nu2, sigma2)
    return first_gaussian + second_gaussian


def _normpdf(x: np.array or float, chi: float, nu: float, std: float):
    """
    Helper function for `double_gaussian`, computes the PDF of a
    gaussian function.

    Args:
        x (np.array or float): Set of input values to evaluate.
        chi (float): Height of the gaussian.
        nu (float): Average value for the gaussian.
        std (float): Standard deviation of the gaussian.

    Returns:
        result (float): PDF value.
    """

    var = std ** 2
    num = np.exp(- (x - nu) ** 2 / (2 * var))
    return chi * num


if __name__ == '__main__':
    help(parse_residues)
    help(coordination_scorer)
    help(discrete_score)
    help(gaussian_score)
    help(bimodal)
